"""
pyrad.io.trajectory
===================
Trajectory class implementation for reading trajectory file.
Converting to different coordinate systems.
.. autosummary::
    :toctree: generated/
    Trajectory
    _Radar_Trajectory
"""

import sys
import re
import datetime
import locale
from warnings import warn
from copy import deepcopy

import numpy as np
import pandas as pd

import pyart

from ..io.read_data_sensor import read_lightning, read_trt_traj_data
from ..io.read_data_sensor import read_trt_thundertracking_traj_data


class Trajectory:
    """
    A class for reading and handling trajectory data from a file.
    Attributes
    ----------
    filename : str
        Path and name of the trajectory definition file
    starttime : datetime
        Start time of trajectory processing.
    endtime : datetime
        End time of trajectory processing.
    trajtype : str
    Type of trajectory. Can be 'plane' or 'lightning'
    time_vector : Array of datetime objects
        Array containing the trajectory time samples
    wgs84_lat_deg : Array of floats
        WGS84 latitude samples in deg
    wgs84_lon_deg : Array of floats
        WGS84 longitude samples in deg
    wgs84_alt_m : Array of floats
        WGS84 altitude samples in m
    nsamples : int
    Number of samples in the trajectory
    _swiss_grid_done : Bool
        Indicates that convertion to Swiss coordinates has been performed
    swiss_chy, swiss_chx, swiss_chh : Array of floats
        Swiss coordinates in m
    radar_list : list
          List of radars for which trajectories are going to be computed
    flashnr : int
          For 'lightning' only. Number of flash for which trajectory data
          is going to be computed. If 0 all all flashes are going to be
          considered.
    time_in_flash : array of floats
          For 'lightning' only. Time within flash (sec)
    flashnr_vec : array of ints
          For 'lightning' only. Flash number of each data sample
    dBm : array of floats
          For 'lightning' only. Lightning power (dBm)
    Methods:
    --------
    add_radar : Add a radar
    calculate_velocities : Computes velocities
    get_start_time : Return time of first trajectory sample
    get_end_time : Return time of last trajectory sample
    get_samples_in_period : Get indices of samples within period
    _convert_traj_to_swissgrid : convert data from WGS84 to Swiss coordinates
    _read_traj : Read plane trajectory from file
    _read_traj_flores : Read FLORAKO plane trajectory from file
    _read_traj_trt : Read TRT trajectory from file
    _read_traj_lightning : Read lightning trajectory from file
    _get_total_seconds : Get the total time of the trajectory in seconds
    """

    def __init__(
        self, filename, starttime=None, endtime=None, trajtype="plane", flashnr=0
    ):
        """
        Initalize the object.
        Parameters
        ----------
        filename : str
            Filename containing the trajectory samples.
        starttime : datetime
            Start time of trajectory processing. If not given, use
            the time of the first trajectory sample.
        endtime : datetime
            End time of trajectory processing. If not given, use
            the time of the last trajectory sample.
        trajtype : str
            type of trajectory. Can be plane or lightning
        flashnr : int
            If type of trajectory is lightning, the flash number to check the
            trajectory. 0 means all flash numbers included
        """

        self.filename = filename
        self.starttime = starttime
        self.endtime = endtime
        self.trajtype = trajtype

        self.time_vector = np.array([], dtype=datetime.datetime)
        self.wgs84_lat_deg = np.array([], dtype=float)
        self.wgs84_lon_deg = np.array([], dtype=float)
        self.wgs84_alt_m = np.array([], dtype=float)
        self.nsamples = None

        self._swiss_grid_done = False
        self.swiss_chy = None
        self.swiss_chx = None
        self.swiss_chh = None

        self.radar_list = []

        if self.trajtype == "lightning":
            self.flashnr = flashnr
            self.time_in_flash = np.array([], dtype=datetime.datetime)
            self.dBm = np.array([], dtype=float)
            self.flashnr_vec = np.array([], dtype=float)
        elif self.trajtype == "trt":
            self.cell_contour = np.array([])

        try:
            if self.trajtype == "lightning":
                self._read_traj_lightning(flashnr)
            elif self.trajtype == "trt":
                self._read_traj_trt()
            elif self.trajtype == "plane_flores":
                self._read_traj_flores()
            else:
                self._read_traj()

        except Exception as ee:
            warn(str(ee))
            raise Exception(
                "ERROR: Could not load trajectory data from file '"
                + filename
                + "' into Trajectory object"
            )

    def add_radar(self, radar):
        """
        Add the coordinates (WGS84 longitude, latitude and non WGS84 altitude)
        of a radar to the radar_list.
        Parameters
        ----------
        radar : pyart radar object
            containing the radar coordinates
        Return
        ------
        Radar object
        """

        # Check if radar location is already in the radar list
        for rad in self.radar_list:
            if rad.location_is_equal(
                radar.latitude["data"][0],
                radar.longitude["data"][0],
                radar.altitude["data"][0],
            ):
                # warn("WARNING: Tried to add the same radar twice to the"
                #      " radar list")
                return rad

        rad = _Radar_Trajectory(
            radar.latitude["data"][0],
            radar.longitude["data"][0],
            radar.altitude["data"][0],
        )
        self.radar_list.append(rad)

        # Convert trajectory WGS84 points to polara radar coordinates
        rad.convert_radpos_to_swissgrid()

        self._convert_traj_to_swissgrid()

        # Note: Earth curvature not considered yet!

        (rvec, azvec, elvec) = pyart.core.cartesian_to_antenna(
            self.swiss_chy - rad.ch_y,
            self.swiss_chx - rad.ch_x,
            self.swiss_chh - rad.ch_alt,
        )

        rad.assign_trajectory(elvec, azvec, rvec)

        return rad

    def calculate_velocities(self, radar):
        """
        Calculate velocities.
        """

        if not radar.traj_assigned:
            raise Exception("ERROR: No trajectory assigned to radar object")

        dt_secs = np.vectorize(self._get_total_seconds)
        dt = dt_secs(self.time_vector[2:] - self.time_vector[:-2])

        v_r = np.empty(self.nsamples, dtype=float)
        v_r[0] = v_r[-1] = np.nan
        v_r[1:-1] = (radar.range_vec[2:] - radar.range_vec[:-2]) / dt

        v_az = np.empty(self.nsamples, dtype=float)
        v_az[0] = v_az[-1] = np.nan
        daz = radar.azimuth_vec[2:] - radar.azimuth_vec[:-2]
        daz[daz > 180.0] -= 360.0
        daz[daz < -180.0] += 360.0
        v_az[1:-1] = daz / dt

        v_el = np.empty(self.nsamples, dtype=float)
        v_el[0] = v_el[-1] = np.nan
        v_el[1:-1] = (radar.elevation_vec[2:] - radar.elevation_vec[:-2]) / dt

        v_abs = np.empty(self.nsamples, dtype=float)
        v_abs[0] = v_abs[-1] = np.nan
        self._convert_traj_to_swissgrid()
        dx = self.swiss_chy[2:] - self.swiss_chy[:-2]
        dy = self.swiss_chx[2:] - self.swiss_chx[:-2]
        dz = self.swiss_chh[2:] - self.swiss_chh[:-2]
        v_abs[1:-1] = np.sqrt(dx**2 + dy**2 + dz**2) / dt

        radar.assign_velocity_vecs(v_abs, v_r, v_el, v_az)

    def get_samples_in_period(self, start=None, end=None):
        """ "
        Get indices of samples of the trajectory within given time
        period.
        """

        if (start is None) and (end is None):
            raise Exception("ERROR: Either start or end must be defined")
        if start is None:
            return np.where(self.time_vector < end)
        if end is None:
            return np.where(self.time_vector >= start)
        return np.where((self.time_vector >= start) & (self.time_vector < end))

    def get_start_time(self):
        """
        Get time of first trajectory sample.
        Return
        ------
        datetime object
        """

        return self.time_vector[0]

    def get_end_time(self):
        """
        Get time of last trajectory sample.
        Return
        ------
        datetime object
        """

        return self.time_vector[-1]

    def _convert_traj_to_swissgrid(self):
        """
        Convert trajectory samples from WGS84 to Swiss CH1903 coordinates
        """

        if self._swiss_grid_done:
            return

        (
            self.swiss_chy,
            self.swiss_chx,
            self.swiss_chh,
        ) = pyart.core.wgs84_to_swissCH1903(
            self.wgs84_lon_deg, self.wgs84_lat_deg, self.wgs84_alt_m
        )

        self._swiss_grid_done = True

    def _read_traj(self):
        """
        Read trajectory from file
        File format
        -----------
        Comment symbol: '#'
        Columns:
        1. Day (UTC) format: DD-MMM-YYYY
        2. Seconds since midnight (UTC)
        3. WGS84 latitude [radians]
        4. WGS84 longitude [radians]
        5. WGS84 altitude [m]
        """

        # check if the file can be read
        try:
            tfile = open(self.filename, "r")
        except Exception:
            raise Exception(
                "ERROR: Could not find|open trajectory file '" + self.filename + "'"
            )

        repat = re.compile(
            "(\\d+\\-[A-Za-z]+\\-\\d+)\\s+([\\d\\.]+)\\s+"
            "([\\-\\d\\.]+)\\s+([\\-\\d\\.]+)\\s+([\\-\\d\\.]+)"
        )

        try:
            loc_set = False
            loc = locale.getlocale()  # get current locale
            if loc[0] != "en_US":
                try:
                    locale.setlocale(locale.LC_ALL, ("en_US", "UTF-8"))
                except Exception as ee:
                    raise Exception(f"ERROR: Cannot set local 'en_US': {ee}")
                loc_set = True

            recording_started = True
            if self.starttime is not None:
                recording_started = False
            recording_check_stop = False
            if self.endtime is not None:
                recording_check_stop = True

            for line in tfile:
                line = line.strip()

                if not line:
                    continue

                # ignore comments
                if line.startswith("#"):
                    continue

                line = line.partition("#")[0]  # Remove comments
                line = line.strip()

                mm = repat.match(line)
                if not mm:
                    print(
                        f"WARNING: Format error in trajectory file"
                        f" '{self.filename}' on line '{line}'",
                        file=sys.stderr,
                    )
                    continue

                # Get time stamp
                try:
                    sday = datetime.datetime.strptime(mm.group(1), "%d-%b-%Y")
                except Exception as ee:
                    print(
                        datetime.datetime.now(datetime.timezone.utc).strftime(
                            "%d-%b-%Y"
                        )
                    )
                    raise Exception(
                        f"ERROR: Format error in traj file '{self.filename}' "
                        f"on line '{line}' ({str(ee)})"
                    )

                sday += datetime.timedelta(seconds=float(mm.group(2)))

                if not recording_started:
                    if sday < self.starttime:
                        continue
                    recording_started = True

                if recording_check_stop:
                    if sday > self.endtime:
                        break

                self.time_vector = np.append(self.time_vector, [sday])

                self.wgs84_lat_deg = np.append(
                    self.wgs84_lat_deg, [float(mm.group(3)) * 180.0 / np.pi]
                )
                self.wgs84_lon_deg = np.append(
                    self.wgs84_lon_deg, [float(mm.group(4)) * 180.0 / np.pi]
                )
                self.wgs84_alt_m = np.append(self.wgs84_alt_m, [float(mm.group(5))])
        except Exception:
            raise
        finally:
            tfile.close()
            if loc_set:
                locale.setlocale(locale.LC_ALL, loc)  # restore saved locale

        self.nsamples = len(self.time_vector)

    def _read_traj_flores(self):
        """
        Read trajectory from FLORAKO file
        File format
        -----------
        Column/Variable Contents, Units and Description:
        01: UTCDate      Year-Month-Day                 Date of epoch or feature (UTC time)
        02: UTCTime      HH:MM:SS.SS                    Time of epoch or feature - Receiver time frame (UTC)
        03: Latitude     Decimal Degrees (signed)       North/South Geographic coordinate
        04: Longitude    Decimal Degrees (signed)       East/West Geographic coordinate
        05: H-MSL        Metres                         Height above the geoid
        06: H-Ell        Metres                         Height above the current ellipsoid
        07: Undulation   Metres                         Height of the ellipsoid above the geoid
        08: PDOP                                        Position Dilution of Precision, which is a measure of X, Y, Z position geometry
        09: HDOP                                        Horizontal Position Dilution of Precision, which is a measure of X, Y position geometry
        10: VDOP                                        Vertical Position Dilution of Precision, which is a measure of Z position geometry
        11: DopRms                                      Root mean square of all L1 Doppler measurement residuals
        """
        # check if the file can be read
        try:
            tfile = open(self.filename, "r")
        except Exception:
            raise Exception(
                "ERROR: Could not find|open trajectory file '" + self.filename + "'"
            )
        tfile.close()

        try:
            data = pd.read_csv(self.filename, skiprows=28, delim_whitespace=True)
            data = data.drop(0)
            times = [
                datetime.datetime.strptime(d + "_" + h + "0000", "%Y-%m-%d_%H:%M:%S.%f")
                for d, h in zip(data["UTCDate"], data["UTCTime"])
            ]

            self.time_vector = np.array(times)
            self.wgs84_lat_deg = np.array(pd.to_numeric(data["Latitude"]))
            self.wgs84_lon_deg = np.array(pd.to_numeric(data["Longitude"]))
            self.wgs84_alt_m = np.array(pd.to_numeric(data["H-MSL"]))

        except Exception:
            raise

        self.nsamples = len(self.time_vector)

    def _read_traj_lightning(self, flashnr=0):
        """
        Read trajectory from lightning file
        File format
        -----------
        Columns:
        1. flashnr (0=noise)
        2. Seconds since midnight (UTC)
        2. Time in flash (s)
        3. WGS84 latitude [deg]
        4. WGS84 longitude [deg]
        5. WGS84 altitude [m]
        6. Power [dBm]
        Parameters
        ----------
        flashnr : int
            the flash number to keep. If 0 data from all flashes will be kept
        """
        flashnr_vec, time, time_in_flash, lat, lon, alt, dBm = read_lightning(
            self.filename
        )

        if flashnr_vec is None:
            raise Exception(
                "ERROR: Could not find|open trajectory file '" + self.filename + "'"
            )

        recording_started = True
        if self.starttime is not None:
            recording_started = False
        recording_check_stop = False
        if self.endtime is not None:
            recording_check_stop = True

        if flashnr > 0:
            flashnr_vec_aux = deepcopy(flashnr_vec)
            flashnr_vec = flashnr_vec[flashnr_vec_aux == flashnr]
            time = time[flashnr_vec_aux == flashnr]
            time_in_flash = time_in_flash[flashnr_vec_aux == flashnr]
            lat = lat[flashnr_vec_aux == flashnr]
            lon = lon[flashnr_vec_aux == flashnr]
            alt = alt[flashnr_vec_aux == flashnr]
            dBm = dBm[flashnr_vec_aux == flashnr]

        for i, dBm_val in enumerate(dBm):
            if not recording_started:
                if time[i] < self.starttime:
                    continue
                recording_started = True

            if recording_check_stop:
                if time[i] > self.endtime:
                    break

            self.flashnr_vec = np.append(self.flashnr_vec, [flashnr_vec[i]])
            self.time_vector = np.append(self.time_vector, [time[i]])
            self.time_in_flash = np.append(self.time_in_flash, [time_in_flash[i]])

            self.wgs84_lat_deg = np.append(self.wgs84_lat_deg, [lat[i]])
            self.wgs84_lon_deg = np.append(self.wgs84_lon_deg, [lon[i]])
            self.wgs84_alt_m = np.append(self.wgs84_alt_m, [alt[i]])

            self.dBm = np.append(self.dBm, [dBm_val])

        self.nsamples = len(self.time_vector)

    def _read_traj_trt(self):
        """
        Read trajectory from TRT file
        File format
        -----------
        Columns:
        1. traj_ID
        2. yyyymmddHHMM (UTC)
        2. lon (deg)
        3. lat [deg]
        4. ell_L
        5. ell_S
        6. ell_or
        7. area
        8. vel_x
        9. vel_y
        10. det
        11. RANKr
        12. CG_n
        13. CG_p
        14. CG
        15. CG_percent_p
        16. ET45
        17. ET45m
        18. ET15
        19. ET15m
        20. VIL
        21. maxH
        22. maxHm
        23. POH
        24. RANK
        25. Dvel_x
        26. Dvel_y
        27. cell_contours
        Parameters
        ----------
        """
        if "_tt.trt" in self.filename:
            (
                traj_ID,
                _,
                yyyymmddHHMM,
                _,
                _,
                _,
                lon,
                lat,
                _,
                _,
                _,
                _,
                _,
                _,
                _,
                _,
                _,
                _,
                _,
                _,
                _,
                _,
                _,
                _,
                _,
                _,
                _,
                _,
                _,
                _,
                _,
                cell_contours,
            ) = read_trt_thundertracking_traj_data(self.filename)

            if traj_ID is None:
                raise Exception(
                    "ERROR: Could not find|open trajectory file '" + self.filename + "'"
                )

            valid = np.logical_not(np.ma.getmaskarray(yyyymmddHHMM))
            yyyymmddHHMM = yyyymmddHHMM[valid]
            traj_ID = traj_ID[valid]
            lon = lon[valid]
            lat = lat[valid]
            cell_contours = cell_contours[valid]

            if traj_ID.size == 0:
                raise Exception(
                    "ERROR: No valid data in trajectory file '" + self.filename + "'"
                )

        else:
            (
                traj_ID,
                yyyymmddHHMM,
                lon,
                lat,
                _,
                _,
                _,
                _,
                _,
                _,
                _,
                _,
                _,
                _,
                _,
                _,
                _,
                _,
                _,
                _,
                _,
                _,
                _,
                _,
                _,
                _,
                _,
                cell_contours,
            ) = read_trt_traj_data(self.filename)

            if traj_ID is None:
                raise Exception(
                    "ERROR: Could not find|open trajectory file '" + self.filename + "'"
                )

        recording_started = True
        if self.starttime is not None:
            recording_started = False
        recording_check_stop = False
        if self.endtime is not None:
            recording_check_stop = True

        for i, cell_contour in enumerate(cell_contours):
            if not recording_started:
                if yyyymmddHHMM[i] < self.starttime:
                    continue
                recording_started = True

            if recording_check_stop:
                if yyyymmddHHMM[i] > self.endtime:
                    break

            self.time_vector = np.append(self.time_vector, [yyyymmddHHMM[i]])

            self.wgs84_lat_deg = np.append(self.wgs84_lat_deg, [lat[i]])
            self.wgs84_lon_deg = np.append(self.wgs84_lon_deg, [lon[i]])
            self.wgs84_alt_m = np.append(self.wgs84_alt_m, 0.0)

            self.cell_contour = np.append(self.cell_contour, [cell_contour])

        self.nsamples = len(self.time_vector)

    def _get_total_seconds(self, x):
        """Return total seconds of timedelta object"""
        return x.total_seconds()


class _Radar_Trajectory:
    """
    A class for holding the trajectory data assigned to a radar.
    Attributes
    ----------
    latitude : float
       WGS84 radar latitude [deg]
    longitude : float
       WGS84 radar longitude [deg]
    altitude : float
       radar altitude [m] (non WGS84)
    ch_y, ch_x, ch_alt : float
       radar coordinates in swiss CH1903 coordinates
    elevation_vec : float list
       Elevation values of the trajectory samples
    azimuth_vec : float list
       Azimuth values of the trajectory samples
    range_vec : float list
       Range values of the trajectory samples
    v_abs, v_r, v_el, v_az : array-like
       Velocity vectors of the absolute [m/s], radial [m/s], elevation [deg/s]
       and azimuth [deg/s] velocities
    Methods:
    --------
    location_is_equal
    assign_trajectory
    convert_radpos_to_swissgrid
    assign_velocity_vecs
    """

    def __init__(self, lat, lon, alt):
        """
        Initalize the object.
        Parameters
        ----------
        lat, lon , alt : radar location coordinates
        nsamps : number of samples
        """

        # radar position in lat, lon
        self.latitude = lat
        self.longitude = lon
        self.altitude = alt

        # radar position in swiss CH1903 coordinates
        self._swiss_coords_done = False
        self.ch_y = None
        self.ch_x = None
        self.ch_alt = None

        # trajectory in polar radar coordinates
        self.traj_assigned = False
        self.elevation_vec = None
        self.azimuth_vec = None
        self.range_vec = None

        # Velocity vectors
        self._velocity_vecs_assigned = False
        self.v_abs = None
        self.v_r = None
        self.v_el = None
        self.v_az = None

    def location_is_equal(self, lat, lon, alt):
        """
        Check if the given coordinates are the same.
        Parameters
        ----------
        lat, lon , alt : radar location coordinates
        Return
        ------
        True if the radar location is equal, False otherwise
        """

        lat_tol = 0.002  # [deg]
        lon_tol = 0.002  # [deg]
        alt_tol = 2.0  # [m]

        if (
            (np.abs(self.latitude - lat) > lat_tol)
            or (np.abs(self.longitude - lon) > lon_tol)
            or (np.abs(self.altitude - alt) > alt_tol)
        ):
            return False

        return True

    def assign_trajectory(self, el, az, rr):
        """
        Assign a trajectory to the radar in polar radar
        coordinates.
        Parameters
        ----------
        el, az, rr : array-like
           elevation, azimuth and range vector
        """

        if self.traj_assigned:
            warn("WARNING: Trajectory already assigned")
            return

        self.elevation_vec = el
        self.azimuth_vec = az
        self.range_vec = rr
        self.traj_assigned = True

    def convert_radpos_to_swissgrid(self):
        """
        Convert the radar location (in WGS84 coordinates) to
        swiss CH1903 coordinates.
        """

        if self._swiss_coords_done:
            return

        (self.ch_y, self.ch_x, self.ch_alt) = pyart.core.wgs84_to_swissCH1903(
            self.longitude, self.latitude, self.altitude, no_altitude_transform=True
        )
        self._swiss_coords_done = True

    def assign_velocity_vecs(self, v_abs, v_r, v_el, v_az):
        """
        Assign velocity vectors to the radar.
        """

        if self._velocity_vecs_assigned:
            warn("WARNING: Trajectory velocity vectors already assigned")
            return

        self.v_abs = v_abs
        self.v_r = v_r
        self.v_az = v_az
        self.v_el = v_el
        self._velocity_vecs_assigned = True



"""
pyrad.proc.process_phase
======================================

Functions for PhiDP and KDP processing and attenuation correction

.. autosummary::
    :toctree: generated/

    process_correct_phidp0
    process_smooth_phidp_single_window
    process_smooth_phidp_double_window
    process_kdp_leastsquare_single_window
    process_kdp_leastsquare_double_window
    process_phidp_kdp_Vulpiani
    process_phidp_kdp_Kalman
    process_phidp_kdp_Maesaka
    process_phidp_kdp_lp
    process_selfconsistency_kdp_phidp
    process_selfconsistency_bias
    process_attenuation

"""

import warnings

from copy import deepcopy
from warnings import warn

import numpy as np

import pyart

from ..io.io_aux import get_datatype_fields
from ..io.read_data_sensor import read_fzl_igra

# Ignore warning in process_phidp_kdp_Maesaka
warnings.filterwarnings(
    "ignore",
    message=".*'partition' will ignore the 'mask' of the MaskedArray.*",
    category=UserWarning,
)


def process_correct_phidp0(procstatus, dscfg, radar_list=None):
    """
    corrects phidp of the system phase

    Parameters
    ----------
    procstatus : int
        Processing status: 0 initializing, 1 processing volume,
        2 post-processing
    dscfg : dictionary of dictionaries
        data set configuration. Accepted Configuration Keywords::

        datatype : list of string. Dataset keyword
            The input data types, must contain,
            "dBZ" or "dBZc", and,
            "PhiDP" or "PhiDPc" or "uPhiDP"
        rmin : float. Dataset keyword
            The minimum range where to look for valid data [m]. Default 1000.
        rmax : float. Dataset keyword
            The maximum range where to look for valid data [m]. Default 50000.
        rcell : float. Dataset keyword
            The length of a continuous cell to consider it valid precip [m].
            Default 1000.
        Zmin : float. Dataset keyword
            The minimum reflectivity [dBZ]. Default 20.
        Zmax : float. Dataset keyword
            The maximum reflectivity [dBZ]. Default 40.
    radar_list : list of Radar objects
        Optional. list of radar objects

    Returns
    -------
    new_dataset : dict
        dictionary containing the output field "PhiDPc"
    ind_rad : int
        radar index

    """

    if procstatus != 1:
        return None, None

    for datatypedescr in dscfg["datatype"]:
        radarnr, _, datatype, _, _ = get_datatype_fields(datatypedescr)
        if datatype == "dBZ":
            refl_field = "reflectivity"
        if datatype == "dBZc":
            refl_field = "corrected_reflectivity"
        if datatype == "PhiDP":
            psidp_field = "differential_phase"
        if datatype == "PhiDPc":
            psidp_field = "corrected_differential_phase"
        if datatype == "uPhiDP":
            psidp_field = "uncorrected_differential_phase"

    ind_rad = int(radarnr[5:8]) - 1
    if radar_list[ind_rad] is None:
        warn("No valid radar")
        return None, None
    radar = radar_list[ind_rad]

    if (refl_field not in radar.fields) or (psidp_field not in radar.fields):
        warn("Unable to correct PhiDP system offset. Missing data")
        return None, None

    # user defined parameters
    rmin = dscfg.get("rmin", 1000.0)
    rmax = dscfg.get("rmax", 50000.0)
    rcell = dscfg.get("rcell", 1000.0)
    zmin = dscfg.get("Zmin", 20.0)
    zmax = dscfg.get("Zmax", 40.0)

    ind_rmin = np.where(radar.range["data"] > rmin)[0][0]
    ind_rmax = np.where(radar.range["data"] < rmax)[0][-1]
    r_res = radar.range["data"][1] - radar.range["data"][0]
    min_rcons = int(rcell / r_res)

    if psidp_field.startswith("corrected_"):
        phidp_field = psidp_field
    elif psidp_field.startswith("uncorrected_"):
        phidp_field = psidp_field.replace("uncorrected_", "corrected_", 1)
    else:
        phidp_field = "corrected_" + psidp_field

    try:
        phidp = pyart.correct.correct_sys_phase(
            radar,
            ind_rmin=ind_rmin,
            ind_rmax=ind_rmax,
            min_rcons=min_rcons,
            zmin=zmin,
            zmax=zmax,
            psidp_field=psidp_field,
            refl_field=refl_field,
            phidp_field=phidp_field,
        )
    except AttributeError:
        warn("Could not find correct_sys_phase function...")
        warn("Skipping system phase correction...")
        warn("Please use PyART-MCH to get this functionality")
        phidp = radar.fields[phidp_field]

    # prepare for exit
    new_dataset = {"radar_out": deepcopy(radar)}
    new_dataset["radar_out"].fields = dict()
    new_dataset["radar_out"].add_field(phidp_field, phidp)

    return new_dataset, ind_rad


def process_smooth_phidp_single_window(procstatus, dscfg, radar_list=None):
    """
    corrects phidp of the system phase and smoothes it using one window

    Parameters
    ----------
    procstatus : int
        Processing status: 0 initializing, 1 processing volume,
        2 post-processing
    dscfg : dictionary of dictionaries
        data set configuration. Accepted Configuration Keywords::

        datatype : list of string. Dataset keyword
            The input data types, must contain,
            "dBZ" or "dBZc", and,
            "PhiDP" or "PhiDPc" or "uPhiDP"
        rmin : float. Dataset keyword
            The minimum range where to look for valid data [m]. Default 1000.
        rmax : float. Dataset keyword
            The maximum range where to look for valid data [m]. Default 50000.
        rcell : float. Dataset keyword
            The length of a continuous cell to consider it valid precip [m].
            Default 1000.
        rwind : float. Dataset keyword
            The length of the smoothing window [m]. Default 6000.
        Zmin : float. Dataset keyword
            The minimum reflectivity [dBZ]. Default 20.
        Zmax : float. Dataset keyword
            The maximum reflectivity [dBZ]. Default 40.
    radar_list : list of Radar objects
        Optional. list of radar objects

    Returns
    -------
    new_dataset : dict
        dictionary containing the output field "PhiDPc"
    ind_rad : int
        radar index

    """

    if procstatus != 1:
        return None, None

    for datatypedescr in dscfg["datatype"]:
        radarnr, _, datatype, _, _ = get_datatype_fields(datatypedescr)
        if datatype == "dBZ":
            refl_field = "reflectivity"
        if datatype == "dBZc":
            refl_field = "corrected_reflectivity"
        if datatype == "PhiDP":
            psidp_field = "differential_phase"
        if datatype == "PhiDPc":
            psidp_field = "corrected_differential_phase"
        if datatype == "uPhiDP":
            psidp_field = "uncorrected_differential_phase"

    ind_rad = int(radarnr[5:8]) - 1
    if radar_list[ind_rad] is None:
        warn("No valid radar")
        return None, None
    radar = radar_list[ind_rad]

    if (refl_field not in radar.fields) or (psidp_field not in radar.fields):
        warn("Unable to smooth PhiDP. Missing data")
        return None, None

    # user defined parameters
    rmin = dscfg.get("rmin", 1000.0)
    rmax = dscfg.get("rmax", 50000.0)
    rcell = dscfg.get("rcell", 1000.0)
    zmin = dscfg.get("Zmin", 20.0)
    zmax = dscfg.get("Zmax", 40.0)
    rwind = dscfg.get("rwind", 6000.0)

    ind_rmin = np.where(radar.range["data"] > rmin)[0][0]
    ind_rmax = np.where(radar.range["data"] < rmax)[0][-1]
    r_res = radar.range["data"][1] - radar.range["data"][0]
    min_rcons = int(rcell / r_res)
    wind_len = int(rwind / r_res)
    min_valid = int(wind_len / 2 + 1)

    if psidp_field.startswith("corrected_"):
        phidp_field = psidp_field
    elif psidp_field.startswith("uncorrected_"):
        phidp_field = psidp_field.replace("uncorrected_", "corrected_", 1)
    else:
        phidp_field = "corrected_" + psidp_field

    phidp = pyart.correct.smooth_phidp_single_window(
        radar,
        ind_rmin=ind_rmin,
        ind_rmax=ind_rmax,
        min_rcons=min_rcons,
        zmin=zmin,
        zmax=zmax,
        wind_len=wind_len,
        min_valid=min_valid,
        psidp_field=psidp_field,
        refl_field=refl_field,
        phidp_field=phidp_field,
    )

    # prepare for exit
    new_dataset = {"radar_out": deepcopy(radar)}
    new_dataset["radar_out"].fields = dict()
    new_dataset["radar_out"].add_field(phidp_field, phidp)

    return new_dataset, ind_rad


def process_smooth_phidp_double_window(procstatus, dscfg, radar_list=None):
    """
    corrects phidp of the system phase and smoothes it using one window

    Parameters
    ----------
    procstatus : int
        Processing status: 0 initializing, 1 processing volume,
        2 post-processing
    dscfg : dictionary of dictionaries
        data set configuration. Accepted Configuration Keywords::

        datatype : list of string. Dataset keyword
            The input data types, must contain,
            "dBZ" or "dBZc", and,
            "PhiDP" or "PhiDPc" or "uPhiDP"
        rmin : float. Dataset keyword
            The minimum range where to look for valid data [m]
        rmax : float. Dataset keyword
            The maximum range where to look for valid data [m]
        rcell : float. Dataset keyword
            The length of a continuous cell to consider it valid precip [m]
        rwinds : float. Dataset keyword
            The length of the short smoothing window [m]
        rwindl : float. Dataset keyword
            The length of the long smoothing window [m]
        Zmin : float. Dataset keyword
            The minimum reflectivity [dBZ]
        Zmax : float. Dataset keyword
            The maximum reflectivity [dBZ]
        Zthr : float. Dataset keyword
            The threshold defining wich smoothed data to used [dBZ]
    radar_list : list of Radar objects
        Optional. list of radar objects

    Returns
    -------
    new_dataset : dict
        dictionary containing the output field "PhiDPc"
    ind_rad : int
        radar index

    """

    if procstatus != 1:
        return None, None

    for datatypedescr in dscfg["datatype"]:
        radarnr, _, datatype, _, _ = get_datatype_fields(datatypedescr)
        if datatype == "dBZ":
            refl_field = "reflectivity"
        if datatype == "dBZc":
            refl_field = "corrected_reflectivity"
        if datatype == "PhiDP":
            psidp_field = "differential_phase"
        if datatype == "PhiDPc":
            psidp_field = "corrected_differential_phase"
        if datatype == "uPhiDP":
            psidp_field = "uncorrected_differential_phase"

    ind_rad = int(radarnr[5:8]) - 1
    if radar_list[ind_rad] is None:
        warn("No valid radar")
        return None, None
    radar = radar_list[ind_rad]

    if (refl_field not in radar.fields) or (psidp_field not in radar.fields):
        warn("Unable to smooth PhiDP. Missing data")
        return None, None

    # user defined parameters
    rmin = dscfg.get("rmin", 1000.0)
    rmax = dscfg.get("rmax", 50000.0)
    rcell = dscfg.get("rcell", 1000.0)
    zmin = dscfg.get("Zmin", 20.0)
    zmax = dscfg.get("Zmax", 40.0)
    rwinds = dscfg.get("rwinds", 2000.0)
    rwindl = dscfg.get("rwindl", 6000.0)
    zthr = dscfg.get("Zthr", 40.0)

    ind_rmin = np.where(radar.range["data"] > rmin)[0][0]
    ind_rmax = np.where(radar.range["data"] < rmax)[0][-1]
    r_res = radar.range["data"][1] - radar.range["data"][0]
    min_rcons = int(rcell / r_res)
    swind_len = int(rwinds / r_res)
    smin_valid = int(swind_len / 2 + 1)
    lwind_len = int(rwindl / r_res)
    lmin_valid = int(lwind_len / 2 + 1)

    if psidp_field.startswith("corrected_"):
        phidp_field = psidp_field
    elif psidp_field.startswith("uncorrected_"):
        phidp_field = psidp_field.replace("uncorrected_", "corrected_", 1)
    else:
        phidp_field = "corrected_" + psidp_field

    phidp = pyart.correct.smooth_phidp_double_window(
        radar,
        ind_rmin=ind_rmin,
        ind_rmax=ind_rmax,
        min_rcons=min_rcons,
        zmin=zmin,
        zmax=zmax,
        swind_len=swind_len,
        smin_valid=smin_valid,
        lwind_len=lwind_len,
        lmin_valid=lmin_valid,
        zthr=zthr,
        psidp_field=psidp_field,
        refl_field=refl_field,
        phidp_field=phidp_field,
    )

    # prepare for exit
    new_dataset = {"radar_out": deepcopy(radar)}
    new_dataset["radar_out"].fields = dict()
    new_dataset["radar_out"].add_field(phidp_field, phidp)

    return new_dataset, ind_rad


def process_phidp_kdp_Maesaka(procstatus, dscfg, radar_list=None):
    """
    Estimates PhiDP and KDP using the method by Maesaka. This method only
    retrieves data in rain (i.e. below the melting layer)

    Parameters
    ----------
    procstatus : int
        Processing status: 0 initializing, 1 processing volume,
        2 post-processing
    dscfg : dictionary of dictionaries
        data set configuration. Accepted Configuration Keywords::

        datatype : list of string. Dataset keyword
            The input data types, must contain,
            "dBZ" or "dBZc", and,
            "PhiDP" or "PhiDPc" or "uPhiDP", and
            "TEMP" or "H_ISO0" (Optional)
        rmin : float. Dataset keyword
            The minimum range where to look for valid data [m]. Default 1000.
        rmax : float. Dataset keyword
            The maximum range where to look for valid data [m]. Default 50000.
        rcell : float. Dataset keyword
            The length of a continuous cell to consider it valid precip [m].
            Default 1000.
        Zmin : float. Dataset keyword
            The minimum reflectivity [dBZ]. Default 20
        Zmax : float. Dataset keyword
            The maximum reflectivity [dBZ]. Default 40
        fzl : float. Dataset keyword
            The freezing level height [m]. Default 2000.
        sounding : str. Dataset keyword
            The nearest radiosounding WMO code (5 int digits). It will be used
            to compute the freezing level, if no temperature field name is
            specified, if the temperature field isin the radar object or if no
            freezing_level is explicitely defined.
        ml_thickness : float. Dataset keyword
            The melting layer thickness in meters. Default 700.
        beamwidth : float. Dataset keyword
            the antenna beamwidth [deg]. If None that of the keys
            radar_beam_width_h or radar_beam_width_v in attribute
            instrument_parameters of the radar object will be used. If the key
            or the attribute are not present the beamwidth will be set to None
    radar_list : list of Radar objects
        Optional. list of radar objects

    Returns
    -------
    new_dataset : dict
        dictionary containing the output field "PhiDPc" and "KDPc"
    ind_rad : int
        radar index

    """

    if procstatus != 1:
        return None, None

    temp_field = None
    iso0_field = None
    for datatypedescr in dscfg["datatype"]:
        radarnr, _, datatype, _, _ = get_datatype_fields(datatypedescr)
        if datatype == "PhiDP":
            psidp_field = "differential_phase"
        if datatype == "PhiDPc":
            psidp_field = "corrected_differential_phase"
        if datatype == "uPhiDP":
            psidp_field = "uncorrected_differential_phase"
        if datatype == "dBZ":
            refl_field = "reflectivity"
        if datatype == "dBZc":
            refl_field = "corrected_reflectivity"
        if datatype == "TEMP":
            temp_field = "temperature"
        if datatype == "H_ISO0":
            iso0_field = "height_over_iso0"

    ind_rad = int(radarnr[5:8]) - 1
    if radar_list[ind_rad] is None:
        warn("No valid radar")
        return None, None
    radar = radar_list[ind_rad]

    if (refl_field not in radar.fields) or (psidp_field not in radar.fields):
        warn(
            "Unable to retrieve PhiDP KDP using the Maesaka approach. " + "Missing data"
        )
        return None, None

    # user defined parameters
    rmin = dscfg.get("rmin", 1000.0)
    rmax = dscfg.get("rmax", 50000.0)
    rcell = dscfg.get("rcell", 1000.0)
    zmin = dscfg.get("Zmin", 20.0)
    zmax = dscfg.get("Zmax", 40.0)
    thickness = dscfg.get("ml_thickness", 700.0)

    fzl = None
    temp_ref = "fixed_fzl"
    # determine which freezing level reference
    if temp_field is not None:
        if temp_field in radar.fields:
            temp_ref = "temperature"
        else:
            warn(
                "COSMO temperature field not available. "
                + "Using fixed freezing level height to determine liquid phase"
            )
    elif iso0_field is not None:
        if iso0_field in radar.fields:
            temp_ref = "height_over_iso0"
        else:
            warn(
                "Height over iso0 field not available. "
                + "Using fixed freezing level height to determine liquid phase"
            )
    else:
        # determine freezing level height if necessary
        temp_ref = "fixed_fzl"
        if "fzl" in dscfg:
            fzl = dscfg["fzl"]
        elif "sounding" in dscfg:
            sounding_code = dscfg["sounding"]
            t0 = pyart.util.datetime_utils.datetime_from_radar(radar)
            fzl = read_fzl_igra(sounding_code, t0, dscfg=dscfg)
        else:
            warn("Freezing level height not defined. Using default " + str(fzl) + " m")
            fzl = 2000

    phidp_field = "corrected_differential_phase"
    kdp_field = "corrected_specific_differential_phase"

    radar_aux = deepcopy(radar)

    # correct PhiDP0
    ind_rmin = np.where(radar_aux.range["data"] > rmin)[0][0]
    ind_rmax = np.where(radar_aux.range["data"] < rmax)[0][-1]
    r_res = radar_aux.range["data"][1] - radar_aux.range["data"][0]
    min_rcons = int(rcell / r_res)

    try:
        phidp = pyart.correct.correct_sys_phase(
            radar_aux,
            ind_rmin=ind_rmin,
            ind_rmax=ind_rmax,
            min_rcons=min_rcons,
            zmin=zmin,
            zmax=zmax,
            psidp_field=psidp_field,
            refl_field=refl_field,
            phidp_field=phidp_field,
        )
    except AttributeError:
        warn("Could not find correct_sys_phase function...")
        warn("Skipping system phase correction...")
        warn("Please use PyART-MCH to get this functionality")
        phidp = radar_aux.fields[phidp_field]

    # filter out data in an above the melting layer
    mask = np.ma.getmaskarray(phidp["data"])

    beamwidth = dscfg.get("beamwidth", None)
    if beamwidth is None:
        if radar.instrument_parameters is not None:
            if "radar_beam_width_h" in radar.instrument_parameters:
                beamwidth = radar.instrument_parameters["radar_beam_width_h"]["data"][0]
            elif "radar_beam_width_v" in radar.instrument_parameters:
                beamwidth = radar.instrument_parameters["radar_beam_width_v"]["data"][0]
    if beamwidth is None:
        warn("Antenna beam width unknown.")

    try:
        mask_fzl, _ = pyart.correct.get_mask_fzl(
            radar_aux,
            fzl=fzl,
            doc=15,
            min_temp=0.0,
            max_h_iso0=0.0,
            thickness=thickness,
            beamwidth=beamwidth,
            temp_field=temp_field,
            iso0_field=iso0_field,
            temp_ref=temp_ref,
        )
        mask = np.logical_or(mask, mask_fzl)
    except AttributeError:
        warn("Masking above FZL is only available in PyART-MCH")
        warn("Please use this library instead of ARM's version")

    # filter out data with invalid reflectivity
    mask_refl = np.ma.getmaskarray(radar_aux.fields[refl_field]["data"])
    mask = np.logical_or(mask, mask_refl)

    phidp["data"] = np.ma.masked_where(mask, phidp["data"])
    fill_value = pyart.config.get_fillvalue()
    radar_aux.add_field(phidp_field, phidp, replace_existing=True)

    # the return data is not a masked array
    kdp, phidpf, _ = pyart.retrieve.kdp_proc.kdp_maesaka(
        radar_aux,
        gatefilter=None,
        method="cg",
        backscatter=None,
        Clpf=1.0,
        length_scale=None,
        first_guess=0.01,
        finite_order="low",
        fill_value=fill_value,
        psidp_field=phidp_field,
        kdp_field=kdp_field,
        phidp_field=phidp_field,
    )

    kdp["data"] = np.ma.masked_where(mask, kdp["data"])
    phidpf["data"] = np.ma.masked_where(mask, phidpf["data"])

    # prepare for exit
    new_dataset = {"radar_out": deepcopy(radar_aux)}
    new_dataset["radar_out"].fields = dict()
    new_dataset["radar_out"].add_field(phidp_field, phidpf)
    new_dataset["radar_out"].add_field(kdp_field, kdp)

    return new_dataset, ind_rad


def process_phidp_kdp_lp(procstatus, dscfg, radar_list=None):
    """
    Estimates PhiDP and KDP using a linear programming algorithm.

    This method only retrieves data in rain (i.e. below the melting layer).

    The method has 3 steps: PhiDP unfolding (including clutter suppression),
    Processing PhiDP using the LP algorithm, Obtaining KDP by convoluting
    PhiDP with a Sobel window.

    Required inputs for the LP algorithm are PsiDP and reflectivity.

    RhoHV and SNR are used for the clutter suppression in the PhiDP unfolding
    step (note that SNR is used instead of Normalized Coherent Power used by
    the original algorithm). If they are not provided a gate_filter based on
    the values of reflectivity is used instead.

    Freezing level height can be retrieved from iso-0 or temperature fields,
    from radio sounding or set by the user.

    Parameters
    ----------
    procstatus : int
        Processing status: 0 initializing, 1 processing volume,
        2 post-processing
    dscfg : dictionary of dictionaries
        data set configuration. Accepted Configuration Keywords::

        datatype : list of string. Dataset keyword
            The input data types, must contain,
            "dBZ" or "dBZc", and,
            "PhiDP" or "PhiDPc" or "uPhiDP", and,
            "RhoHV" or "RhoHVc", (Optional, used when min_rhv is specified) and,
            "SNRh" (Optional, used when min_snr is specified), and
            "TEMP" or "H_ISO0" (Optional)
        fzl : float. Dataset keyword
            The freezing level height [m]. Default 2000.
        sounding : str. Dataset keyword
            The nearest radiosounding WMO code (5 int digits). It will be used
            to compute the freezing level, if no temperature field name is
            specified, if the temperature field is not in the radar object or
            if no freezing_level is explicitely defined.
        ml_thickness : float. Dataset keyword
            The melting layer thickness in meters. Default 700.
        beamwidth : float. Dataset keyword
            the antenna beamwidth [deg]. If None that of the keys
            radar_beam_width_h or radar_beam_width_v in attribute
            instrument_parameters of the radar object will be used. If the key
            or the attribute are not present the beamwidth will be set to None
        LP_solver : str.
            The LP solver to use. Can be pyglpk, cvxopt, cylp, cylp_mp.
            Default cvxopt
        proc : int
            Number of worker processes. Only used when LP_solver is cylp_mp.
            Default 1
        min_snr, min_rhv : float
            Minimum SNR and RhoHV. Used to filter out non-meteorological
            echoes when performing the unfolding of the differential phase.
            Default 10 and 0.6
        sys_phase : float
            Default system differential phase offset in deg. Default 0.
        overide_sys_phase : bool
            If True the dynamic sys_phase not computed and the default
            sys_phase is used. If false a dynamic sys_phase is computed. If
            no dynamic value is found the default sys_phase is used.
            Default False
        first_gate_sysp : int
            First gate to use when determining the system differential phase
            offset. Default 0
        nowrap : int or None
            Gate number where to begin the phase unwrapping. None will unwrap
            from gate 0. Default None
        ncpts : int
            Minimum number of continuous valid PhiDP points. Segments below
            this number or starting at a gate below this number are going to
            be excluded from the unfolding. Default 2.
        z_bias : float
            reflectivity bias. Default 0 dBZ
        low_z, high_z : float
            mininum and maximum reflectivity values. Values beyond this range
            are going to be set to this range limits. The modified
            reflectivity is used in the LP algorithm. Default 10 and 53 dBZ
        min_phidp : float
            minimum differential phase. PhiDP values below this threshold are
            going to be set to the threshold values. The modified PhiDP is
            used in the LP algorithm. Default 0.1 deg
        doc : int
            Number of gates to doc at the end of the ray. Used in the LP
            algorithm. Default 0
        self_const : float
            selfconsistency factor. Used in the LP algorithm. Default 60000
        coef : float
            Exponent linking Z to KDP in selfconsistency. Used in the LP
            algorithm. kdp = (10**(0.1z))**coef. Default 0.914
        window_len : int
            Length of Sobel Window applied to PhiDP field before computing
            KDP. Default 35

    radar_list : list of Radar objects
        Optional. list of radar objects

    Returns
    -------
    new_dataset : dict
        dictionary containing the output field "PhiDPc" and "KDPc"
    ind_rad : int
        radar index

    """

    if procstatus != 1:
        return None, None

    temp_field = None
    iso0_field = None
    rhv_field = None
    snr_field = None
    for datatypedescr in dscfg["datatype"]:
        radarnr, _, datatype, _, _ = get_datatype_fields(datatypedescr)
        if datatype == "PhiDP":
            psidp_field = "differential_phase"
        if datatype == "PhiDPc":
            psidp_field = "corrected_differential_phase"
        if datatype == "uPhiDP":
            psidp_field = "uncorrected_differential_phase"
        if datatype == "dBZ":
            refl_field = "reflectivity"
        if datatype == "dBZc":
            refl_field = "corrected_reflectivity"
        if datatype == "RhoHVc":
            rhv_field = "corrected_cross_correlation_ratio"
        if datatype == "RhoHV":
            rhv_field = "cross_correlation_ratio"
        if datatype == "SNRh":
            snr_field = "signal_to_noise_ratio_hh"
        if datatype == "TEMP":
            temp_field = "temperature"
        if datatype == "H_ISO0":
            iso0_field = "height_over_iso0"

    ind_rad = int(radarnr[5:8]) - 1
    if radar_list[ind_rad] is None:
        warn("No valid radar")
        return None, None
    radar = radar_list[ind_rad]

    if (refl_field not in radar.fields) or (psidp_field not in radar.fields):
        warn(
            "Unable to retrieve PhiDP and KDP using the LP approach. "
            + "Missing reflectivity and/or PsiDP data"
        )
        return None, None

    fzl = None
    temp_ref = "fixed_fzl"
    # determine which freezing level reference
    if temp_field is not None:
        if temp_field in radar.fields:
            temp_ref = "temperature"
        else:
            warn(
                "COSMO temperature field not available. "
                + "Using fixed freezing level height to determine liquid phase"
            )
    elif iso0_field is not None:
        if iso0_field in radar.fields:
            temp_ref = "height_over_iso0"
        else:
            warn(
                "Height over iso0 field not available. "
                + "Using fixed freezing level height to determine liquid phase"
            )
    else:
        # determine freezing level height if necessary
        temp_ref = "fixed_fzl"
        if "fzl" in dscfg:
            fzl = dscfg["fzl"]
        elif "sounding" in dscfg:
            sounding_code = dscfg["sounding"]
            t0 = pyart.util.datetime_utils.datetime_from_radar(radar)
            fzl = read_fzl_igra(sounding_code, t0, dscfg=dscfg)
        else:
            fzl = 2000
            warn("Freezing level height not defined. Using default " + str(fzl) + " m")

    radar_aux = deepcopy(radar)

    # user config
    LP_solver = dscfg.get("LP_solver", "cvxopt")
    thickness = dscfg.get("ml_thickness", 700.0)
    offset = -dscfg.get("z_bias", 0.0)
    self_const = dscfg.get("self_const", 60000.0)
    low_z = dscfg.get("low_z", 10.0)
    high_z = dscfg.get("high_z", 53.0)
    min_phidp = dscfg.get("min_phidp", 0.01)
    min_ncp = dscfg.get("min_snr", 10.0)
    min_rhv = dscfg.get("min_rhv", 0.6)
    sys_phase = dscfg.get("sys_phase", 0.0)
    first_gate_sysp = dscfg.get("first_gate_sysp", 0)
    ncpts = dscfg.get("ncpts", 2)
    overide_sys_phase = dscfg.get("overide_sys_phase", False)
    nowrap = dscfg.get("nowrap", None)
    window_len = dscfg.get("window_len", 35)
    proc = dscfg.get("proc", 1)
    coef = dscfg.get("coef", 0.914)
    doc = dscfg.get("doc", 0)

    # filter out data in an above the melting layer
    mask = np.ma.getmaskarray(radar_aux.fields[psidp_field]["data"])
    beamwidth = dscfg.get("beamwidth", None)
    if beamwidth is None:
        if radar.instrument_parameters is not None:
            if "radar_beam_width_h" in radar.instrument_parameters:
                beamwidth = radar.instrument_parameters["radar_beam_width_h"]["data"][0]
            elif "radar_beam_width_v" in radar.instrument_parameters:
                beamwidth = radar.instrument_parameters["radar_beam_width_v"]["data"][0]
    if beamwidth is None:
        warn("Antenna beam width unknown.")

    mask_fzl, _ = pyart.correct.get_mask_fzl(
        radar_aux,
        fzl=fzl,
        doc=15,
        min_temp=0.0,
        max_h_iso0=0.0,
        thickness=thickness,
        beamwidth=beamwidth,
        temp_field=temp_field,
        iso0_field=iso0_field,
        temp_ref=temp_ref,
    )
    mask = np.logical_or(mask, mask_fzl)

    # filter out data with invalid reflectivity
    mask_refl = np.ma.getmaskarray(radar_aux.fields[refl_field]["data"])
    mask = np.logical_or(mask, mask_refl)

    radar_aux.fields[psidp_field]["data"] = np.ma.masked_where(
        mask, radar_aux.fields[psidp_field]["data"]
    )

    phidp_field = "corrected_differential_phase"
    kdp_field = "corrected_specific_differential_phase"

    # fzl is None meaning temperature fields could be used:
    # set fzl to an arbitrary high value
    if fzl is None:
        fzl = 5000.0

    if rhv_field is not None and snr_field is not None:
        phidp, kdp = pyart.correct.phase_proc_lp(
            radar_aux,
            offset,
            debug=False,
            self_const=self_const,
            low_z=low_z,
            high_z=high_z,
            min_phidp=min_phidp,
            min_ncp=min_ncp,
            min_rhv=min_rhv,
            fzl=fzl,
            sys_phase=sys_phase,
            ncpts=ncpts,
            overide_sys_phase=overide_sys_phase,
            nowrap=nowrap,
            really_verbose=False,
            LP_solver=LP_solver,
            refl_field=refl_field,
            ncp_field=snr_field,
            rhv_field=rhv_field,
            phidp_field=psidp_field,
            kdp_field=kdp_field,
            unf_field=phidp_field,
            window_len=window_len,
            proc=proc,
            coef=coef,
        )
    else:
        if not overide_sys_phase:
            sys_phase = None

        gatefilter = pyart.filters.GateFilter(radar_aux)
        gatefilter.exclude_masked(psidp_field)

        phidp, kdp = pyart.correct.phase_proc_lp_gf(
            radar_aux,
            offset=offset,
            gatefilter=gatefilter,
            debug=False,
            self_const=self_const,
            low_z=low_z,
            high_z=high_z,
            min_phidp=min_phidp,
            fzl=fzl,
            system_phase=sys_phase,
            ncpts=ncpts,
            first_gate_sysp=first_gate_sysp,
            nowrap=nowrap,
            really_verbose=False,
            LP_solver=LP_solver,
            refl_field=refl_field,
            phidp_field=psidp_field,
            kdp_field=kdp_field,
            unf_field=phidp_field,
            window_len=window_len,
            proc=proc,
            coef=coef,
            doc=doc,
        )

    kdp["data"] = np.ma.masked_where(mask, kdp["data"])
    phidp["data"] = np.ma.masked_where(mask, phidp["data"])

    # prepare for exit
    new_dataset = {"radar_out": deepcopy(radar)}
    new_dataset["radar_out"].fields = dict()
    new_dataset["radar_out"].add_field(phidp_field, phidp)
    new_dataset["radar_out"].add_field(kdp_field, kdp)

    return new_dataset, ind_rad


def process_kdp_leastsquare_single_window(procstatus, dscfg, radar_list=None):
    """
    Computes specific differential phase using a piecewise least square method

    Parameters
    ----------
    procstatus : int
        Processing status: 0 initializing, 1 processing volume,
        2 post-processing
    dscfg : dictionary of dictionaries
        data set configuration. Accepted Configuration Keywords::

        datatype : list of string. Dataset keyword
            The input data types, must contain,
            "PhiDP" or "PhiDPc" or "uPhiDP"
        rwind : float. Dataset keyword
            The length of the segment for the least square method [m].
            Default 6000.
        vectorize : bool. Dataset keyword
            Whether to vectorize the KDP processing. Default false
    radar_list : list of Radar objects
        Optional. list of radar objects

    Returns
    -------
    new_dataset : dict
        dictionary containing the output field "KDPc"
    ind_rad : int
        radar index

    """
    if procstatus != 1:
        return None, None

    for datatypedescr in dscfg["datatype"]:
        radarnr, _, datatype, _, _ = get_datatype_fields(datatypedescr)
        if datatype == "PhiDP":
            phidp_field = "differential_phase"
        if datatype == "PhiDPc":
            phidp_field = "corrected_differential_phase"
        if datatype == "uPhiDP":
            phidp_field = "uncorrected_differential_phase"

    ind_rad = int(radarnr[5:8]) - 1
    if radar_list[ind_rad] is None:
        warn("No valid radar")
        return None, None
    radar = radar_list[ind_rad]

    if phidp_field not in radar.fields:
        warn("Unable to retrieve KDP from PhiDP using least square. " + "Missing data")
        return None, None

    # user defined parameters
    rwind = dscfg.get("rwind", 6000.0)
    vectorize = dscfg.get("vectorize", False)

    r_res = radar.range["data"][1] - radar.range["data"][0]
    wind_len = int(rwind / r_res)
    min_valid = int(wind_len / 2 + 1)
    kdp_field = "corrected_specific_differential_phase"

    kdp = pyart.retrieve.kdp_leastsquare_single_window(
        radar,
        wind_len=wind_len,
        min_valid=min_valid,
        phidp_field=phidp_field,
        kdp_field=kdp_field,
        vectorize=vectorize,
    )

    # prepare for exit
    new_dataset = {"radar_out": deepcopy(radar)}
    new_dataset["radar_out"].fields = dict()
    new_dataset["radar_out"].add_field(kdp_field, kdp)

    return new_dataset, ind_rad


def process_kdp_leastsquare_double_window(procstatus, dscfg, radar_list=None):
    """
    Computes specific differential phase using a piecewise least square method

    Parameters
    ----------
    procstatus : int
        Processing status: 0 initializing, 1 processing volume,
        2 post-processing
    dscfg : dictionary of dictionaries
        data set configuration. Accepted Configuration Keywords::

            The input data types, must contain,
            "PhiDP" or "PhiDPc" or "uPhiDP", and,
            "dBZ" or "dBZc"
        rwinds : float. Dataset keyword
            The length of the short segment for the least square method [m].
            Default 2000.
        rwindl : float. Dataset keyword
            The length of the long segment for the least square method [m].
            Default 6000.
        Zthr : float. Dataset keyword
            The threshold defining which estimated data to use [dBZ]
        vectorize : Bool. Dataset keyword
            Whether to vectorize the KDP processing. Default false
    radar_list : list of Radar objects
        Optional. list of radar objects

    Returns
    -------
    new_dataset : dict
        dictionary containing the output field "KDPc"
    ind_rad : int
        radar index

    """
    if procstatus != 1:
        return None, None

    for datatypedescr in dscfg["datatype"]:
        radarnr, _, datatype, _, _ = get_datatype_fields(datatypedescr)
        if datatype == "PhiDP":
            phidp_field = "differential_phase"
        if datatype == "PhiDPc":
            phidp_field = "corrected_differential_phase"
        if datatype == "uPhiDP":
            phidp_field = "uncorrected_differential_phase"
        if datatype == "dBZ":
            refl_field = "reflectivity"
        if datatype == "dBZc":
            refl_field = "corrected_reflectivity"

    ind_rad = int(radarnr[5:8]) - 1
    if radar_list[ind_rad] is None:
        warn("No valid radar")
        return None, None
    radar = radar_list[ind_rad]

    if (phidp_field not in radar.fields) or (refl_field not in radar.fields):
        warn("Unable to retrieve KDP from PhiDP using least square. " + "Missing data")
        return None, None

    # user defined parameters
    rwinds = dscfg.get("rwinds", 2000.0)
    rwindl = dscfg.get("rwindl", 6000.0)
    zthr = dscfg.get("Zthr", 40.0)
    vectorize = dscfg.get("vectorize", False)

    r_res = radar.range["data"][1] - radar.range["data"][0]
    swind_len = int(rwinds / r_res)
    smin_valid = int(swind_len / 2 + 1)
    lwind_len = int(rwindl / r_res)
    lmin_valid = int(lwind_len / 2 + 1)

    kdp_field = "corrected_specific_differential_phase"

    kdp = pyart.retrieve.kdp_leastsquare_double_window(
        radar,
        swind_len=swind_len,
        smin_valid=smin_valid,
        lwind_len=lwind_len,
        lmin_valid=lmin_valid,
        zthr=zthr,
        phidp_field=phidp_field,
        refl_field=refl_field,
        kdp_field=kdp_field,
        vectorize=vectorize,
    )

    # prepare for exit
    new_dataset = {"radar_out": deepcopy(radar)}
    new_dataset["radar_out"].fields = dict()
    new_dataset["radar_out"].add_field(kdp_field, kdp)

    return new_dataset, ind_rad


def process_phidp_kdp_Vulpiani(procstatus, dscfg, radar_list=None):
    """
    Computes specific differential phase and differential phase using the
    method developed by Vulpiani et al.
    The data is assumed to be clutter free and monotonous

    Parameters
    ----------
    procstatus : int
        Processing status: 0 initializing, 1 processing volume,
        2 post-processing
    dscfg : dictionary of dictionaries
        data set configuration. Accepted Configuration Keywords::

        datatype : list of string. Dataset keyword
            The input data types, must contain,
            "PhiDP" or "PhiDPc" or "uPhiDP"
        rwind : float. Dataset keyword
            The length of the segment [m]. Default 2000.
        n_iter : int. Dataset keyword
            number of iterations. Default 3.
        interp : boolean. Dataset keyword
            if set non valid values are interpolated using neighbouring valid
            values. Default 0 (False)
        parallel : boolean. Dataset keyword
            if set use parallel computing. Default 1 (True)
        get_phidp : boolean. Datset keyword
            if set the PhiDP computed by integrating the resultant KDP is
            added to the radar field. Default 0 (False)
        frequency : float. Dataset keyword
            the radar frequency [Hz]. If None that of the key
            frequency in attribute instrument_parameters of the radar
            object will be used. If the key or the attribute are not present
            it will be assumed that the radar is C band
    radar_list : list of Radar objects
        Optional. list of radar objects

    Returns
    -------
    new_dataset : dict
        dictionary containing the output field "PhiDPc" and "KDPc"
    ind_rad : int
        radar index

    """
    if procstatus != 1:
        return None, None

    for datatypedescr in dscfg["datatype"]:
        radarnr, _, datatype, _, _ = get_datatype_fields(datatypedescr)
        if datatype == "PhiDP":
            phidp_field = "differential_phase"
        if datatype == "PhiDPc":
            phidp_field = "corrected_differential_phase"
        if datatype == "uPhiDP":
            phidp_field = "uncorrected_differential_phase"

    ind_rad = int(radarnr[5:8]) - 1
    if radar_list[ind_rad] is None:
        warn("No valid radar")
        return None, None
    radar = radar_list[ind_rad]

    if phidp_field not in radar.fields:
        warn("Unable to retrieve KDP from PhiDP using least square. " + "Missing data")
        return None, None

    # user defined parameters
    n_iter = dscfg.get("n_iter", 3)
    rwind = dscfg.get("rwind", 2000.0)
    interp = dscfg.get("interp", 0)
    parallel = dscfg.get("parallel", 1)
    get_phidp = dscfg.get("get_phidp", 0)

    # window length (must be even)
    r_res = radar.range["data"][1] - radar.range["data"][0]
    wind_len = int(rwind / r_res)
    if wind_len % 2 == 1:
        wind_len += 1

    # get band
    freq = dscfg.get("frequency", None)
    if freq is None:
        if (
            radar.instrument_parameters is not None
            and "frequency" in radar.instrument_parameters
        ):
            freq = radar.instrument_parameters["frequency"]["data"][0]

    band = "C"
    if freq is None:
        warn("Radar frequency unknown. Default C band will be applied")
    else:
        band = pyart.retrieve.get_freq_band(freq)
        if band is None:
            if freq < 2e9:
                band = "S"
            elif freq > 12e9:
                band = "X"
            warn(
                "Radar frequency out of range. Valid bands are S, C or X. "
                + band
                + " band will be applied"
            )

    kdp_field = "corrected_specific_differential_phase"
    phidpr_field = "corrected_differential_phase"

    kdp_dict, phidpr_dict = pyart.retrieve.kdp_vulpiani(
        radar,
        gatefilter=None,
        fill_value=None,
        psidp_field=phidp_field,
        kdp_field=kdp_field,
        phidp_field=phidpr_field,
        band=band,
        windsize=wind_len,
        n_iter=n_iter,
        interp=interp,
        prefilter_psidp=False,
        filter_opt=None,
        parallel=parallel,
    )

    # prepare for exit
    new_dataset = {"radar_out": deepcopy(radar)}
    new_dataset["radar_out"].fields = dict()
    new_dataset["radar_out"].add_field(kdp_field, kdp_dict)
    if get_phidp:
        new_dataset["radar_out"].add_field(phidpr_field, phidpr_dict)

    return new_dataset, ind_rad


def process_phidp_kdp_Kalman(procstatus, dscfg, radar_list=None):
    """
    Computes specific differential phase and differential phase using the
    Kalman filter as proposed by Schneebeli et al.
    The data is assumed to be clutter free and continous

    Parameters
    ----------
    procstatus : int
        Processing status: 0 initializing, 1 processing volume,
        2 post-processing
    dscfg : dictionary of dictionaries
        data set configuration. Accepted Configuration Keywords::

        datatype : list of string. Dataset keyword
            The input data types, must contain,
            "PhiDP" or "PhiDPc" or "uPhiDP"
        parallel : boolean. Dataset keyword
            if set use parallel computing
        get_phidp : boolean. Datset keyword
            if set the PhiDP computed by integrating the resultant KDP is
            added to the radar field
        frequency : float. Dataset keyword
            the radar frequency [Hz]. If None that of the key
            frequency in attribute instrument_parameters of the radar
            object will be used. If the key or the attribute are not present
            it will be assumed that the radar is C band
    radar_list : list of Radar objects
        Optional. list of radar objects

    Returns
    -------
    new_dataset : dict
        dictionary containing the output field "PhiDPc" and "KDPc"
    ind_rad : int
        radar index

    """
    if procstatus != 1:
        return None, None

    for datatypedescr in dscfg["datatype"]:
        radarnr, _, datatype, _, _ = get_datatype_fields(datatypedescr)
        if datatype == "PhiDP":
            phidp_field = "differential_phase"
        if datatype == "PhiDPc":
            phidp_field = "corrected_differential_phase"
        if datatype == "uPhiDP":
            phidp_field = "uncorrected_differential_phase"

    ind_rad = int(radarnr[5:8]) - 1
    if radar_list[ind_rad] is None:
        warn("No valid radar")
        return None, None
    radar = radar_list[ind_rad]

    if phidp_field not in radar.fields:
        warn("Unable to retrieve KDP from PhiDP using least square. " + "Missing data")
        return None, None

    # User defined options
    parallel = dscfg.get("parallel", 1)
    get_phidp = dscfg.get("get_phidp", 0)

    # get band
    freq = dscfg.get("frequency", None)
    if freq is None:
        if (
            radar.instrument_parameters is not None
            and "frequency" in radar.instrument_parameters
        ):
            freq = radar.instrument_parameters["frequency"]["data"][0]

    band = "C"
    if freq is None:
        warn("Radar frequency unknown. Default C band will be applied")
    else:
        band = pyart.retrieve.get_freq_band(freq)
        if band is None:
            if freq < 2e9:
                band = "S"
            elif freq > 12e9:
                band = "X"
            warn(
                "Radar frequency out of range. Valid bands are S, C or X. "
                + band
                + " band will be applied"
            )

    kdp_field = "corrected_specific_differential_phase"
    phidpr_field = "corrected_differential_phase"

    kdp_dict, _, phidpr_dict = pyart.retrieve.kdp_schneebeli(
        radar,
        gatefilter=None,
        fill_value=None,
        psidp_field=phidp_field,
        kdp_field=kdp_field,
        phidp_field=phidpr_field,
        band=band,
        rcov=0,
        pcov=0,
        prefilter_psidp=False,
        filter_opt=None,
        parallel=parallel,
    )

    # prepare for exit
    new_dataset = {"radar_out": deepcopy(radar)}
    new_dataset["radar_out"].fields = dict()
    new_dataset["radar_out"].add_field(kdp_field, kdp_dict)
    if get_phidp:
        new_dataset["radar_out"].add_field(phidpr_field, phidpr_dict)

    return new_dataset, ind_rad


def process_attenuation(procstatus, dscfg, radar_list=None):
    """
    Computes specific attenuation and specific differential attenuation using
    the Z-Phi method and corrects reflectivity and differential reflectivity

    Parameters
    ----------
    procstatus : int
        Processing status: 0 initializing, 1 processing volume,
        2 post-processing
    dscfg : dictionary of dictionaries
        data set configuration. Accepted Configuration Keywords::

        datatype : list of string. Dataset keyword
            The input data types, must contain,
            "dBZ" or "dBZc", and,
            "PhiDP" or "PhiDPc", and,
            "ZDR" or "ZDRc", and
            "TEMP" or "H_ISO0" (Optional)
        ATT_METHOD : float. Dataset keyword
            The attenuation estimation method used. One of the following:
            ZPhi, Philin. Default ZPhi
        fzl : float. Dataset keyword
            The default freezing level height. It will be used if no
            temperature field name is specified or the temperature field is
            not in the radar object. Default 2000.
        sounding : str. Dataset keyword
            The nearest radiosounding WMO code (5 int digits). It will be used
            to compute the freezing level, if no temperature field name is
            specified, if the temperature field is in the radar object or if
            no freezing_level is explicitely defined.
    radar_list : list of Radar objects
        Optional. list of radar objects

    Returns
    -------
    new_dataset : dict
        dictionary containing the output fields "Ah" (spec. attenuation),
        "PIA" (path-integrated attenuation) and "dBZc" (corr. refl.)
    ind_rad : int
        radar index

    """

    if procstatus != 1:
        return None, None

    temp = None
    iso0 = None
    zdr = None
    for datatypedescr in dscfg["datatype"]:
        radarnr, _, datatype, _, _ = get_datatype_fields(datatypedescr)
        if datatype == "dBZc":
            refl = "corrected_reflectivity"
        if datatype == "PhiDPc":
            phidp = "corrected_differential_phase"
        if datatype == "ZDRc":
            zdr = "corrected_differential_reflectivity"
        if datatype == "dBZ":
            refl = "reflectivity"
        if datatype == "PhiDP":
            phidp = "differential_phase"
        if datatype == "ZDR":
            zdr = "differential_reflectivity"
        if datatype == "TEMP":
            temp = "temperature"
        if datatype == "H_ISO0":
            iso0 = "height_over_iso0"

    ind_rad = int(radarnr[5:8]) - 1
    if radar_list[ind_rad] is None:
        warn("No valid radar")
        return None, None
    radar = radar_list[ind_rad]

    if phidp not in radar.fields or refl not in radar.fields:
        warn("Unable to compute attenuation. Missing data")
        return None, None

    # determine which freezing level reference
    fzl = None
    temp_ref = "fixed_fzl"
    if temp is not None:
        if temp in radar.fields:
            temp_ref = "temperature"
        else:
            warn(
                "COSMO temperature field not available. "
                + "Using fixed freezing level height to determine liquid phase"
            )
    elif iso0 is not None:
        if iso0 in radar.fields:
            temp_ref = "height_over_iso0"
        else:
            warn(
                "Height over iso0 field not available. "
                + "Using fixed freezing level height to determine liquid phase"
            )
    else:
        # determine freezing level height if necessary
        temp_ref = "fixed_fzl"
        if "fzl" in dscfg:
            fzl = dscfg["fzl"]
        elif "sounding" in dscfg:
            sounding_code = dscfg["sounding"]
            t0 = pyart.util.datetime_utils.datetime_from_radar(radar)
            fzl = read_fzl_igra(sounding_code, t0, dscfg=dscfg)
        else:
            warn("Freezing level height not defined. Using default " + str(fzl) + " m")
            fzl = 2000

    att_method = dscfg.get("ATT_METHOD", "ZPhi")
    if att_method not in ("ZPhi", "Philin"):
        raise ValueError(
            "Unknown attenuation correction method. "
            + "Must be one of the following: [ZPhi, Philin]"
        )

    if att_method == "ZPhi":
        spec_at, pia, cor_z, spec_diff_at, pida, cor_zdr = (
            pyart.correct.calculate_attenuation_zphi(
                radar,
                doc=15,
                fzl=fzl,
                smooth_window_len=0,
                a_coef=None,
                beta=None,
                c=None,
                d=None,
                refl_field=refl,
                phidp_field=phidp,
                zdr_field=zdr,
                temp_field=temp,
                iso0_field=iso0,
                spec_at_field=None,
                pia_field=None,
                corr_refl_field=None,
                spec_diff_at_field=None,
                pida_field=None,
                corr_zdr_field=None,
                temp_ref=temp_ref,
            )
        )
    elif att_method == "Philin":
        spec_at, pia, cor_z, spec_diff_at, pida, cor_zdr = (
            pyart.correct.calculate_attenuation_philinear(
                radar,
                doc=15,
                fzl=fzl,
                pia_coef=None,
                pida_coef=None,
                refl_field=refl,
                phidp_field=phidp,
                zdr_field=zdr,
                temp_field=temp,
                iso0_field=iso0,
                spec_at_field=None,
                pia_field=None,
                corr_refl_field=None,
                spec_diff_at_field=None,
                pida_field=None,
                corr_zdr_field=None,
                temp_ref=temp_ref,
            )
        )

    # prepare for exit
    new_dataset = {"radar_out": deepcopy(radar)}
    new_dataset["radar_out"].fields = dict()

    new_dataset["radar_out"].add_field("specific_attenuation", spec_at)
    new_dataset["radar_out"].add_field("path_integrated_attenuation", pia)
    new_dataset["radar_out"].add_field("corrected_reflectivity", cor_z)

    if (spec_diff_at is not None) and (cor_zdr is not None):
        new_dataset["radar_out"].add_field(
            "specific_differential_attenuation", spec_diff_at
        )
        new_dataset["radar_out"].add_field(
            "path_integrated_differential_attenuation", pida
        )
        new_dataset["radar_out"].add_field(
            "corrected_differential_reflectivity", cor_zdr
        )
    else:
        warn(
            " Specific differential attenuation and attenuation "
            + "corrected differential reflectivity not available"
        )

    return new_dataset, ind_rad
